//////////////////////////////////////////////////////////////////////////////
//
// Blend
//
// A simple utility to "blend" two images with selectable blending modes.
// Based on PixelMath blending blendModes given by Juan Conejero in
// http://pixinsight.com/forum/index.php?topic=2409.0
//
// Mike Reid - 2012 February 25
// Gerald Wechselberger, deutsche Übersetzung, 2012 März
// Hartmut Bornemann, Preview, 2014 April
// Erweitert Gerald Wechselberger, 2015 Mai PixelMath StarMask eingebaut
// Thre=0.07 Std, und SatThre=0.08 Std.
// Erweitert Hartmut Bornemann, 03.05.2015 (Attersee Update)
//
//////////////////////////////////////////////////////////////////////////////

#feature-id    Utilities > Blend
#feature-info  An image blending utility.
#feature-icon  Blend.xpm

#include <pjsr/Color.jsh>
#include <pjsr/Sizer.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/StdDialogCode.jsh>

#define VERSION "3.0.0"
#define TITLE "Blend"

// Shadows clipping point in (normalized) MAD units from the median.
#define DEFAULT_AUTOSTRETCH_SCLIP  -2.80
// Target mean background in the [0,1] range.
#define DEFAULT_AUTOSTRETCH_TBGND   0.25
// Apply the same STF to all nominal channels (true), or treat each channel
// separately (false).
#define DEFAULT_AUTOSTRETCH_CLINK   true

//////////////////////////////////////////////////////////////////////////////
//
// Blend Engine
//
function BlendEngine()
{
   this.topWin    = new ImageWindow();
   this.midWin    = new ImageWindow();
   this.bottomWin = new ImageWindow();

   this.blendModes  = new Array();
   this.blendModes["Normal"] = "IMG2";
   this.blendModes["Darken"] = "Min( IMG1, IMG2 )";
   this.blendModes["Add"] = "IMG1 + IMG2";
   this.blendModes["Subtract Upper-Lower"] = "IMG2 - IMG1";
   this.blendModes["Multiply"] = "IMG1 * IMG2";
   this.blendModes["Divide Upper/Lower"] = "IMG2 / IMG1";
   this.blendModes["Divide by Flat (Upper)"] = "IMG1*MEDD/IMG2";
   this.blendModes["Color Burn"] = "~(~IMG1/ IMG2)";
   this.blendModes["Linear Burn"] = "IMG1 + IMG2 - 1";
   this.blendModes["Aufhellen"] = "Max( IMG1, IMG2 )";
   this.blendModes["Screen (Negativ Multiply)"] = "~(~IMG1 * ~IMG2)";
   this.blendModes["Color Dodge"] = "IMG1 / ~IMG2";
   this.blendModes["Overlay"] = "iif( IMG1 > 0.5, ~(~(2*(IMG1 - 0.5)) * ~IMG2), 2*IMG1*IMG2 )";
   this.blendModes["Soft Light"] = "iif( IMG2 > 0.5, ~(~IMG1 * ~(IMG2 - 0.5)), IMG1*(IMG2 + 0.5) )";
   this.blendModes["Hard Light"] = "iif( IMG2 > 0.5, ~(~IMG1 * ~(2*(IMG2 - 0.5))), 2*IMG1*IMG2 ) ";
   this.blendModes["Vivid Light"] = "iif( IMG2 > 0.5, ~(~IMG1/(IMG2 - 0.5)/2), IMG1/~(2*IMG2) )";
   this.blendModes["Linear Light"] = "iif( IMG2 > 0.5, IMG1 + 2*(IMG2 - 0.5), IMG1 + 2*IMG2 - 1 )";
   this.blendModes["Pin Light"] = "iif( IMG2 > 0.5, Max( IMG1, 2*(IMG2 - 0.5) ), Min( IMG1, 2*IMG2 ) )";
   this.blendModes["Difference"] = "IMG1 -- IMG2";
   this.blendModes["Exclusion"] = "0.5 - 2*(IMG1 - 0.5)*(IMG2 - 0.5)";
   this.blendModes["StarMask"] = "iif((x()==0||x()==w()-1 || y()==0 || y()==h()-1),0,iif((((sqrt((mean(-(pixel($T,x()," +
   "y())-pixel($T,x()-1,y())),(pixel($T,x(),y())-pixel($T,x()+1,y()))))^2+(mean(-(pixel($T,x()," +
   "y())-pixel($T,x(),y()-1)),(pixel($T,x(),y())-pixel($T,x(),y()+1))))^2))>Thre)||($T>SatThre)),1,0))";

   this.createdView = null;
   this.blended     = false;
   this.formula     = "";
   this.inverted    = false;
   this.opacity     = 100;
   this.threshold   = 0.070;
   this.satthre     = 0.8;
   this.whiteMask   = true;
   this.starmask    = false;

   this.SetBlendMode = function (key)
   {
      this.formula = this.blendModes[key];
   }

   this.SetOpacity = function(Opacity)
   {
      this.opacity = Opacity;
   }

   this.SetThreshold = function(Threshold)
   {
      this.threshold = Threshold;
   }

   this.SetSatThre = function(SatThre)
   {
      this.satthre = SatThre;
   }


   this.SetInvertMode = function(Inverted)
   {
      this.inverted = Inverted;
   }

   this.SetWhiteMask = function(WhiteMask)
   {
      this.whiteMask = WhiteMask;
   }

   this.GetFormula = function()
   {
      var scl_img1 = ((100 - this.opacity)/100).toFixed(3);
      var scl_blnd = (this.opacity/100).toFixed(3);

      // principle formula:
      // formula =
      // ( scale_img1 * IMG1 + scale_blend * ( formula ) ) * (MASK) +
      //       (IMG2 * (1-MASK) )

      // inverted mask: MASK = 1 - MASK
      // formula =
      // ( scale_img1 * IMG1 + scale_blend * ( formula ) ) * (1 - MASK) +
      //       (IMG2 * MASK )

      // white mask, i.e. mask == 1
      // formula = scale_img1 * IMG1 + scale_blend * ( formula )
      //

      var Formula;
      if (this.whiteMask)
      {
         Formula =
         "(" +
         scl_img1 + " * IMG1) + " + scl_blnd + " * (" + this.formula + ")";
      }
      else
      {
         if (this.inverted)
         {
            Formula =
            "(" +
                scl_img1 + " * IMG1 + " + scl_blnd + " * (" + this.formula + ")" +
            ")" +
               " * (1 - MASK) + IMG2 * MASK";
         }
         else
         {
            Formula =
            "(" +
                     scl_img1 + " * IMG1 + " + scl_blnd + " * (" + this.formula + ")" +
            ")" +
               " * MASK + IMG2 * (1 - MASK)";
         }
      }
      return Formula;
   }

   this.blendImages = function(intoView, invertMask)
   {
      var formula = this.GetFormula();
      /*
      Console.writeln("Formula");
      Console.writeln(formula);
      Console.writeln("--- *** ---");
      */
      var scl_threshold = this.threshold;
      var scl_satthre = this.satthre;
      var p = new PixelMath();
      with ( p )
      {
         if (this.whiteMask)
         {
         expression =
            "  IMG1 = " + this.bottomWin.mainView.id +
            "; IMG2 = " + this.topWin.mainView.id +
            "; MEDD = med(" + this.topWin.mainView.id + ");" +
            formula;
            symbols = "IMG1, IMG2, MEDD, Thre=" + scl_threshold + ", SatThre="+scl_satthre;
         }
         else
         {
            expression =
               "  MASK = " + this.midWin.mainView.id +
               "; IMG1 = " + this.bottomWin.mainView.id +
               "; IMG2 = " + this.topWin.mainView.id +
               "; MEDD = med(" + this.topWin.mainView.id + ");" +
               formula;
            symbols = "IMG1, IMG2, MASK, MEDD, Thre="+ scl_threshold + ", SatThre="+scl_satthre;
         }
         /*
         Console.writeln("Symbols");
         Console.writeln(symbols);
         Console.writeln("--- *** ---");
         */
         expression1 = "";
         expression2 = "";
         expression3 = "";
         useSingleExpression = true;
         use64BitWorkingImage = false;
         rescale = false;
         rescaleLower = 0.0000000000;
         rescaleUpper = 1.0000000000;
         truncate = true;
         truncateLower = 0.0000000000;
         truncateUpper = 1.0000000000;
         createNewImage = true;  // changed true to false
         newImageId = "blend";
         newImageWidth = 0;
         newImageHeight = 0;
         newImageAlpha = false;
         newImageColorSpace = PixelMath.prototype.SameAsTarget;
         newImageSampleFormat = PixelMath.prototype.SameAsTarget;
         showNewImage = true;
      }
      engine.createdView = null;
      if (! p.executeOn(intoView) )
      {
         engine.blended = false;
         (new MessageBox("PixelMath -  Error", TITLE + " " + VERSION)).execute();
      }
      else
      {
         var w = ImageWindow.activeWindow;
         engine.createdView = w.currentView;
         // Console.writeln("New view created: " + engine.createdView.id);
         engine.blended = true;
      }
      p = null;
   }
}

var engine = new BlendEngine();

//////////////////////////////////////////////////////////////////////////////
//
// Blend Dialog
//
function BlendDialog()
{
   // Add all properties and methods of the core Dialog object to this object.
   this.__base__ = Dialog;
   this.__base__();

   var dlg = this;

   var labelWidth1 = this.font.width( "Overlay Mode:" + 'T' );
   var running     = false;
   this.prvImgW    = null;

   // ------------------------------------------------------------------------
   // Top help panel
   // ------------------------------------------------------------------------

   this.helpLabel = new Label( this );
   with ( this.helpLabel )
   {
      frameStyle = FrameStyle_Box;
      margin = 4;
      readOnly = true;
      wordWrapping = true;
      useRichText = true;

      text = "<p><b>" + TITLE + " v" + VERSION + "</b> &mdash; " +
     "Blend Two Layers Utility.</p>" +
     "<p>Mixes/Overlays two images using PixelMath." +
     " Both images are specified as <b>Upper</b> and <b>Lower</b> Image." +
     " Not all Overlay modes are cummutativ. The ruler limits" +
     " how much from the Upper image is mixed in the Lower image." +
     " This can be repeated as often as you want to find the right Blend percentage." +
     " The mask is used for the result in the Upper image.</p>" +
     " SPECIAL CASE - If you select StarMask as Overlay Mode, a starmask will be created from the Upper image.</p>" +
     " important    - the image has to be selected in Upper AND Lower.   </p>" +
     "<p>Authors: Mike Reid, Gerald Wechselberger, Hartmut Bornemann</p>" +
     "<p>Based on: http://pixinsight.com/forum/index.php?topic=2409.0</p>";
   }

   var tmr = new Timer();
   with (tmr)
   {
      interval = 0.500;
      periodic = true;
      onTimeout = function()
      {
         tmr.stop();
         preview();
      }
   }

   // ------------------------------------------------------------------------
   // GroupBox 'Bildauswahl'
   //
   // Top Layer
   // ------------------------------------------------------------------------

   this.topImageLabel = new Label( this );
   with ( this.topImageLabel ) {
      text = "IMG2 Upper Images:";
      minWidth = labelWidth1;
      textAlignment = TextAlign_Right|TextAlign_VertCenter;
   }

   this.topImageList = new ViewList(this);
   with ( this.topImageList )
   {
      getAll();
      toolTip = "<p>Select Upper image for Overlay</p>";

      onViewSelected = function( view )
      {
         if (view.id.length > 0)
         {
            createPreview(view);
            engine.topWin = view.window;
            preview();
         }
         else
         {
            // view de-selected, create preview from bottom window
            if (dlg.bottomImageList.currentView.id.length > 0)
            {
               // current bottom view is valid
               createPreview(dlg.bottomImageList.currentView);
            }
            else
            {
               if (dlg.prvImgW != null)
               {
                  dlg.prvImgW.forceClose();
               }
            }
         }
      }
   }

   this.topImageSizer = new HorizontalSizer;
   with ( this.topImageSizer )
   {
      spacing = 4;
      add( this.topImageLabel );
      add( this.topImageList );
   }

   // ------------------------------------------------------------------------
   // Middle Layer
   // ------------------------------------------------------------------------

   this.rbWhiteMask = new RadioButton( this );
   with (this.rbWhiteMask)
   {
      checked = true;
      text = "White Mask";
      toolTip = "<p>Sets white Mask</p>";
      onCheck = function( checked )
      {
         engine.SetWhiteMask(checked);
         showFormula();
         dlg.midImageList.enabled = !checked;
         dlg.mask_check.enabled = !checked;
         engine.blended = false;
         preview();
      }
   }

   this.rbPixelMask = new RadioButton( this );
   with (this.rbPixelMask)
   {
      text = "Pixel Mask";
      toolTip = "<p>Select Image-Maske</p>";
   }

   this.radioButtons = new HorizontalSizer;
   with (this.radioButtons)
   {
      add (this.rbWhiteMask );
      add (this.rbPixelMask);
   }

   this.midImageList = new ViewList(this);
   with ( this.midImageList )
   {
      getAll();
      enabled = false;
      minWidth = this.topImageList.width;
      toolTip = this.rbPixelMask.toolTip;

      onViewSelected = function( view )
      {
         if (view.id.length > 0)
         {
            createPreview(view);
            engine.midWin = view.window;
            engine.savedWin = engine.midWin;
            preview();
         }
         else
         {
            dlg.rbWhiteMask.checked = true;
            engine.SetWhiteMask(true);
            showFormula();
            dlg.midImageList.enabled = false;
            dlg.mask_check.enabled = false;
            engine.blended = false;
            preview();
         }
      };
   }

   this.midImageSizer = new HorizontalSizer;
   with ( this.midImageSizer )
   {
      spacing = 4;
      add (this.radioButtons);
      add( this.midImageList );
   }

   // ------------------------------------------------------------------------
   // Bottom Layer
   // ------------------------------------------------------------------------

   this.bottomImageLabel = new Label( this );
   with ( this.bottomImageLabel )
   {
      text = "IMG1 Lower Image:";
      minWidth = labelWidth1;
      textAlignment = TextAlign_Right|TextAlign_VertCenter;
   }

   this.bottomImageList = new ViewList(this);
   with ( this.bottomImageList )
   {
      getAll();
      toolTip = "<p>Select Lower Image for Overlay</p>";

      onViewSelected = function( view )
      {
         if (view.id.length > 0)
         {
            createPreview(view);
            engine.bottomWin = view.window;
            preview();
         }
         else
         {
            // view de-selected, create preview from top window
            if (dlg.topImageList.currentView.id.length > 0)
            {
               // current top view is valid
               createPreview(dlg.topImageList.currentView);
            }
            else
            {
               if (dlg.prvImgW != null)
               {
                  dlg.prvImgW.forceClose();
               }
            }
         }
      };
   }

   this.bottomImageSizer = new HorizontalSizer;
   with ( this.bottomImageSizer )
   {
      spacing = 4;
      add( this.bottomImageLabel );
      add( this.bottomImageList );
   }

   // ------------------------------------------------------------------------
   // Flip Button
   // ------------------------------------------------------------------------

   this.flipImageLabel = new Label( this );
   with ( this.flipImageLabel )
   {
      text = " ";  // alignment trick
      minWidth = labelWidth1;
      textAlignment = TextAlign_Right|TextAlign_VertCenter;
   }

   this.btnFlip = new PushButton( this );
   with ( this.btnFlip )
   {
      //maxWidth = 24;
      text     = "Revert selected Images";
      toolTip = "<p>Replace Upper image with Lower image and vice versa</p>";
      onClick = function()
      {
          var V = dlg.topImageList.currentView;
          dlg.topImageList.currentView = dlg.bottomImageList.currentView;
          dlg.bottomImageList.currentView = V;
          engine.topWin = dlg.topImageList.currentView.window;
          engine.bottomWin = dlg.bottomImageList.currentView.window;
          preview();
      }
   }

   this.flipImageSizer = new HorizontalSizer;
   with ( this.flipImageSizer )
   {
      spacing = 4;
      add( this.flipImageLabel );
      add( this.btnFlip );
   }

   // ------------------------------------------------------------------------
   // GroupBox 'Bildauswahl', vertikale Anordnung der HorizontalSizer-Elemente
   // ------------------------------------------------------------------------

   this.input_GroupBox = new GroupBox( this );
   with ( this.input_GroupBox )
   {
      title = "Select Image";
      sizer = new VerticalSizer;
   }
   with ( this.input_GroupBox.sizer )
   {
      margin = 6;
      spacing = 4;
      add( this.topImageSizer );
      add( this.midImageSizer );
      add( this.bottomImageSizer );
      add( this.flipImageSizer );
   }


   // ------------------------------------------------------------------------
   // GroupBox 'Überblend Optionen'
   //
   // Opacity
   // ------------------------------------------------------------------------

   this.opacity_NC = new NumericControl (this);
   with ( this.opacity_NC )
   {
      enableScientificNotation = false;
      label.text = "Opacity (%):";
      label.minWidth = labelWidth1;
      setRange (0, 100);
      slider.setRange (0, 1000);
      slider.minWidth = 200;
      setPrecision (0);
      setValue (100);
      toolTip = "<p>Opacity Upper Image.</p>";
      onValueUpdated = function (value)
      {
         engine.SetOpacity(value);
         showFormula();
         tmr.start();
      };
   }



   // ------------------------------------------------------------------------
   // GroupBox Starmask Threshold Optionen'
   //
   // Starmask Threshold
   // ------------------------------------------------------------------------

   this.starmask_thre_NC = new NumericControl (this);
   with ( this.starmask_thre_NC )
   {
      enableScientificNotation = false;
      label.text = "Starmask Thre:";
      label.minWidth = labelWidth1;
      setRange (0.001, 1);
      slider.setRange (0.00, 1000);
      slider.minWidth = 200;
      setPrecision (4);
      setValue (0.07);
      toolTip = "<p>Threshold Starmask. Edge -> Star to Background,  lower value -> more stars </p>";
      onValueUpdated = function (value)
      {
         engine.SetThreshold(value);
         showFormula();
//         Console.writeln("Threshold: ");
         tmr.start();
      };
   }



   // ------------------------------------------------------------------------
   // GroupBox Starmask Star Center Saturation Threshold Optionen'
   //
   // Starmask Saturation Threshold
   // ------------------------------------------------------------------------

   this.satthre_NC = new NumericControl (this);
   with ( this.satthre_NC )
   {
      enableScientificNotation = false;
      label.text = "Starmask SatThre:";
      label.minWidth = labelWidth1;
      setRange (0.001, 1);
      slider.setRange (0.00, 1000);
      slider.minWidth = 200;
      setPrecision (4);
      setValue (0.800);
      toolTip = "<p>Star Center Saturation Threshold Starmask. higher value -> darker star center </p>";
      onValueUpdated = function (value)
      {
         engine.SetSatThre(value);
         showFormula();
//         Console.writeln("SatThre: ");
         tmr.start();
      };
   }


   // ------------------------------------------------------------------------
   // Blending mode combo box
   // ------------------------------------------------------------------------

   this.blendModeLabel = new Label( this );
   with ( this.blendModeLabel ) {
      text = "Overlay Mode:";
      minWidth = labelWidth1;
      textAlignment = TextAlign_Right|TextAlign_VertCenter;
   }

   this.blendModeCombo = new ComboBox( this );
   with ( this.blendModeCombo )
   {
      for (var key in engine.blendModes)
      {
         addItem(key);
      }
      engine.SetBlendMode(itemText(0));

      onItemSelected = function( index )
      {
         engine.SetBlendMode(itemText(index));
         showFormula();
         preview();
      }
   }

   // ------------------------------------------------------------------------
   // CheckBox 'Maske invertieren'
   // ------------------------------------------------------------------------

   this.mask_check = new CheckBox( this );
   with (this.mask_check)
   {
      enabled = false;
      text = "Invert Mask";
      textAlignment = TextAlign_Left|TextAlign_VertCenter;
      toolTip = "<p>Invert Mask</p>";
      width = 60;
      onCheck = function( checked )
      {
         engine.SetInvertMode(checked);
         showFormula();
         update();
         preview();
      }
   }

   this.blendModeSizer = new HorizontalSizer;
   with ( this.blendModeSizer )
   {
      spacing = 4;
      add( this.blendModeLabel );
      add( this.blendModeCombo );
      add( this.mask_check );
      addStretch();
   }

   // ------------------------------------------------------------------------
   // GroupBox 'Überblend Optionen', Anordnung Opacity-NumericControl
   // und HorizontalSizer-Element mit Label, ComboBox u. CheckBox
   // ------------------------------------------------------------------------

   this.blend_GroupBox = new GroupBox( this );
   with ( this.blend_GroupBox )
   {
      title = "Overlay Options or StarMask";
      sizer = new VerticalSizer;
   }
   with ( this.blend_GroupBox.sizer )
   {
      margin = 6;
      spacing = 4;
      add( this.opacity_NC );
      add( this.blendModeSizer );
      add( this.starmask_thre_NC );
      add( this.satthre_NC );
   }

   // ------------------------------------------------------------------------
   // Controls im Hauptfenster
   // ------------------------------------------------------------------------

   this.ok_Button = new PushButton( this );
   with ( this.ok_Button )
   {
      text = "Ready";
      onClick = function()
      {
         this.dialog.ok();
      }
   }

   this.activity_Label = new Label(this);
   with (this.activity_Label)
   {
      foregroundColor = backgroundColor;
      text = "running ***";
   }

   this.cbAutoSTF = new CheckBox( this );
   with (this.cbAutoSTF)
   {
      checked = true;
      text = "Auto STF";
      onCheck = function( checked )
      {
         if (checked & (dlg.prvImgW != null))
         {
            update();
            ApplyAutoSTF(dlg.prvImgW.mainView,
                 DEFAULT_AUTOSTRETCH_SCLIP,
                 DEFAULT_AUTOSTRETCH_TBGND,
                 DEFAULT_AUTOSTRETCH_CLINK );
         }
      }
   }

   this.buttons_Sizer = new HorizontalSizer;
   with (this.buttons_Sizer)
   {
      spacing = 6;
      addStretch();
      add( this.cbAutoSTF );
      add( this.activity_Label );
      add( this.ok_Button );
   }

   // ------------------------------------------------------------------------
   // TextBox unten zur Anzeige der Formel
   // ------------------------------------------------------------------------

   this.statusText = new TextBox( this );
   with (this.statusText)
   {
      maxHeight = 100;
      readOnly  = true;
   }

   this.status_Sizer = new HorizontalSizer;
   with (this.status_Sizer)
   {
      add( this.statusText );
   }

   // ------------------------------------------------------------------------
   // Main vertical sizer, vertikale Anordnung von GroupBoxes u. Sizers
   // ------------------------------------------------------------------------

   this.sizer = new VerticalSizer;
   with ( this.sizer )
   {
      margin = 6;
      spacing = 6;
      add( this.helpLabel );
      add( this.input_GroupBox );
      add( this.blend_GroupBox );
      add( this.buttons_Sizer );
      add( this.status_Sizer );
   }

   with ( this )
   {
      windowTitle = TITLE + " Script";
      adjustToContents();
      userResizable = true;
      showFormula();
   }

   //////////////////////////////////////////////////////////////////////////////
   //
   // Preview script extension
   //
   //////////////////////////////////////////////////////////////////////////////

   function createPreview(view)
   {
      if (dlg.prvImgW != null)
      {
         dlg.prvImgW.forceClose();
      }

      with (view.image)
      {
         dlg.prvImgW = new ImageWindow(width,
                                       height,
                                       numberOfChannels,
                                       bitsPerSample,
                                       isReal,
                                       isColor);
      };
      dlg.prvImgW.mainView.beginProcess( UndoFlag_NoSwapFile );
      dlg.prvImgW.mainView.image.apply(view.image);
      dlg.prvImgW.mainView.endProcess();

      if (dlg.cbAutoSTF.checked && engine.createdView != null)
      {
         ApplyAutoSTF(engine.createdView, // dlg.prvImgW.mainView,
                 DEFAULT_AUTOSTRETCH_SCLIP,
                 DEFAULT_AUTOSTRETCH_TBGND,
                 DEFAULT_AUTOSTRETCH_CLINK );
      }
      else
      {
         ApplySTF (dlg.prvImgW.mainView, view.stf);
      }

      dlg.prvImgW.position = new Point(0,0);
      dlg.prvImgW.zoomToFit();
      dlg.prvImgW.show();
      dlg.topImageList.excludedView = dlg.prvImgW.mainView;
      dlg.midImageList.excludedView = dlg.prvImgW.mainView;
      dlg.bottomImageList.excludedView = dlg.prvImgW.mainView;
      dlg.topImageList.reload();
      dlg.midImageList.reload();
      dlg.bottomImageList.reload();
   }

   function preview()
   {
      with (engine)
      {
         if (running) return;
         if (whiteMask)
         {
            if ((topWin.mainView.id.length == 0) ||
                (bottomWin.mainView.id.length == 0))
               return
         }
         else
         {
            if ((topWin.mainView.id.length == 0) ||
                (midWin.mainView.id.length == 0) ||
                (bottomWin.mainView.id.length == 0))
               return;
         }
      }
      running = true;
      dlg.opacity_NC.enabled = false;
      dlg.activity_Label.foregroundColor = Color.RED;
      // run blend engine
      engine.blendImages (dlg.prvImgW.mainView, dlg.mask_check.checked);
      if (engine.blended)
      {
         if (dlg.cbAutoSTF.checked)
         {
            ApplyAutoSTF(engine.createdView,        //   dlg.prvImgW.mainView,
                    DEFAULT_AUTOSTRETCH_SCLIP,
                    DEFAULT_AUTOSTRETCH_TBGND,
                    DEFAULT_AUTOSTRETCH_CLINK );
         }
      }
      else
      {
         (new MessageBox("Blend error, no view created")).execute();
      }
      dlg.activity_Label.foregroundColor = dlg.activity_Label.backgroundColor;
      dlg.opacity_NC.enabled = true;
      running = false;
   }

   function showFormula()
   {
      var formula = engine.GetFormula();
      while (formula.contains("MEDD") && (!engine.topWin.isNull))
      {
         var MEDD = engine.topWin.mainView.image.median().toFixed(4);
         formula = formula.replace("MEDD", MEDD);
      }
      dlg.statusText.text = formula;
   }
}  // end of function BlendDialog()

ApplySTF = function (view, stf)
{
   var low = (stf[0][1] + stf[1][1] + stf[2][1]) / 3;
   var mtf = (stf[0][0] + stf[1][0] + stf[2][0]) / 3;
   var hgh = (stf[0][2] + stf[1][2] + stf[2][2]) / 3;

   if (low > 0 || mtf != 0.5 || hgh != 1) // if not an identity transformation
   {
      // Console.writeln(format("<b>Applying STF to '%ls'</b>:\x1b[38;2;100;100;100m",view.id));
      var HT = new HistogramTransformation;
      HT.H = [
         [  0, 0.5, 1, 0, 1],
         [  0, 0.5, 1, 0, 1],
         [  0, 0.5, 1, 0, 1],
         [low, mtf, hgh, 0, 1],
         [  0, 0.5, 1, 0, 1]
      ];

      HT.executeOn(view, false); // no swap file
      console.write("\x1b[0m");
   }
}

function findMidtonesBalance( v0, v1, eps )
{
   if ( v1 <= 0 )
      return 0;

   if ( v1 >= 1 )
      return 1;

   v0 = Math.range( v0, 0.0, 1.0 );

   if ( eps )
      eps = Math.max( 1.0e-15, eps );
   else
      eps = 5.0e-05;

   var m0, m1;
   if ( v1 < v0 )
   {
      m0 = 0;
      m1 = 0.5;
   }
   else
   {
      m0 = 0.5;
      m1 = 1;
   }

   for ( ;; )
   {
      var m = (m0 + m1)/2;
      var v = Math.mtf( m, v1 );

      if ( Math.abs( v - v0 ) < eps )
         return m;

      if ( v < v0 )
         m1 = m;
      else
         m0 = m;
   }
}

function ApplyAutoSTF( view, shadowsClipping, targetBackground, rgbLinked )
{
   var stf = new ScreenTransferFunction;

   var n = view.image.isColor ? 3 : 1;

   var median = view.computeOrFetchProperty( "Median" );

   var mad = view.computeOrFetchProperty( "MAD" );
   mad.mul( 1.4826 ); // coherent with a normal distribution

   if ( rgbLinked )
   {
      /*
       * Try to find how many channels look as channels of an inverted image.
       * We know a channel has been inverted because the main histogram peak is
       * located over the right-hand half of the histogram. Seems simplistic
       * but this is consistent with astronomical images.
       */
      var invertedChannels = 0;
      for ( var c = 0; c < n; ++c )
         if ( median.at( c ) > 0.5 )
            ++invertedChannels;

      if ( invertedChannels < n )
      {
         /*
          * Noninverted image
          */
         var c0 = 0, m = 0;
         for ( var c = 0; c < n; ++c )
         {
            if ( 1 + mad.at( c ) != 1 )
               c0 += median.at( c ) + shadowsClipping * mad.at( c );
            m  += median.at( c );
         }
         c0 = Math.range( c0/n, 0.0, 1.0 );
         m = Math.mtf( targetBackground, m/n - c0 );

         stf.STF = [ // c0, c1, m, r0, r1
                     [c0, 1, m, 0, 1],
                     [c0, 1, m, 0, 1],
                     [c0, 1, m, 0, 1],
                     [0, 1, 0.5, 0, 1] ];
      }
      else
      {
         /*
          * Inverted image
          */
         var c1 = 0, m = 0;
         for ( var c = 0; c < n; ++c )
         {
            m  += median.at( c );
            if ( 1 + mad.at( c ) != 1 )
               c1 += median.at( c ) - shadowsClipping * mad.at( c );
            else
               c1 += 1;
         }
         c1 = Math.range( c1/n, 0.0, 1.0 );
         m = Math.mtf( c1 - m/n, targetBackground );

         stf.STF = [ // c0, c1, m, r0, r1
                     [0, c1, m, 0, 1],
                     [0, c1, m, 0, 1],
                     [0, c1, m, 0, 1],
                     [0, 1, 0.5, 0, 1] ];
      }
   }
   else
   {
      /*
       * Unlinked RGB channnels: Compute automatic stretch functions for
       * individual RGB channels separately.
       */
      var A = [ // c0, c1, m, r0, r1
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1] ];

      for ( var c = 0; c < n; ++c )
      {
         if ( median.at( c ) < 0.5 )
         {
            /*
             * Noninverted channel
             */
            var c0 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) + shadowsClipping * mad.at( c ), 0.0, 1.0 ) : 0.0;
            var m  = Math.mtf( targetBackground, median.at( c ) - c0 );
            A[c] = [c0, 1, m, 0, 1];
         }
         else
         {
            /*
             * Inverted channel
             */
            var c1 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) - shadowsClipping * mad.at( c ), 0.0, 1.0 ) : 1.0;
            var m  = Math.mtf( c1 - median.at( c ), targetBackground );
            A[c] = [0, c1, m, 0, 1];
         }
      }

      stf.STF = A;
   }
   stf.executeOn( view );
}

// Our dialog inherits all properties and methods from the core Dialog object.
BlendDialog.prototype = new Dialog;

//////////////////////////////////////////////////////////////////////////////
//
// Main script entry point
//
function main()
{
   Console.hide();
   var dialog = new BlendDialog();
   var retVal = dialog.execute();
   if (retVal  == StdDialogCode_Cancel)
   {
      Console.writeln("Cancel");
      if (dialog.prvImgW != null)
      {
         if (!dialog.prvImgW.isNull)
         {
            Console.writeln("Delete preview " + dialog.prvImgW.mainView.id);
            dialog.prvImgW.forceClose();
         }
      }
   }
   else if (retVal  == StdDialogCode_Ok)
   {
      if (!engine.blended)
      {
         if (dialog.prvImgW != null)
         {
            Console.writeln("Delete preview " + dialog.prvImgW.mainView.id);
            dialog.prvImgW.forceClose();
         }
      }
   }
}

main();
